#include <iostream> // pour utiliser std::cout std::endl << >>
#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>

#include "main.h"
#include "LisEtCompileLesShaders.h"


#define TAILLE_HORIZONTALE 800
#define TAILLE_VERTICALE 450

//----------------------------------------
// CLES SUR LA CARTE GRAPHIQUE LIEES AU PROGRAMME GPU DE TESSELLATION

// CLE DU PROGRAMME GPU
static GLuint MonProgrammeGLSL;

// CLE DE LA VARIABLE SOMMET DANS LE PROGRAMME GPU
static GLint MonSommetID;

// CLE DE LA FINESSE DE DECOUPAGE DANS LE PROGRAMME GPU
static GLint FinesseDecoupageID;

//----------------------------------------
// CLES DE LA GEOMETRIE ET DES INDICES DES SOMMETS DES PATCHS SUR LA CARTE GRAPHIQUE
static GLuint MonVAO;
static GLuint MesSommetsID;
static GLuint MesIndicesID;

void InitialisationDuProgrammeGPUEtDesLiensVersCeProgramme()
{
	MonProgrammeGLSL = LisEtCompileLesShaders( 
		"../shader/tessellation.vert.glsl",
		"../shader/tessellation.cont.glsl",
		"../shader/tessellation.eval.glsl",
		"../shader/tessellation.geom.glsl",
		"../shader/tessellation.frag.glsl");

// JE RECHERCHE LES CLES DESIGNANT LES VARIABLES UNIFORMES QUE J UTILISE DANS LES CODES GLSL
// CES VARIABLES SONT DITES UNIFORMES CAR ELLES ONT LA MEME VALEUR POUR TOUS LES SOMMETS/PRIMITIVES/FRAGMENTS.
// LA CHAINE DE CARACTERE A DROITE CONTIENT LE NOM DES VARIABLES DANS LES CODES GLSL
// LE RESULTAT DE LA FONCTION EST LA CLE VERS CETTE VARIABLE GLSL
// CETTE CLE EST UTILISEE PLUS TARD DANS LE CODE C++ POUR TRANSMETTRE UNE VALEUR DU CPU VERS LA VARIABLE GPU
	FinesseDecoupageID = glGetUniformLocation(MonProgrammeGLSL, "u_FinesseDecoupage");
	MonSommetID = glGetAttribLocation(MonProgrammeGLSL, "MonSommetEnEntree");
	std::cout << "MonSommetID = " << MonSommetID << std::endl;
	std::cout << "FinesseDecoupageID = " << FinesseDecoupageID << std::endl;
}


void InitialisationDesDonnees()
{
	glm::vec3 MesSommets[]= {// 6 SOMMETS POUR DEFINIR DEUX PATCHS DE 4 SOMMETS AVEC DEUX SOMMETS EN COMMUN
		glm::vec3(-0.5, -0.5, 0.0),
		glm::vec3( 0.5, -0.5, 0.0),
		glm::vec3(-0.5,  0.0, 0.0),
		glm::vec3( 0.5,  0.0, 0.0),
		glm::vec3(-0.5,  0.5, 0.0),
		glm::vec3( 0.5,  0.5, 0.0)
	}; 
	GLuint MesIndices[]= { 0, 1, 2, 3, 2, 3, 4, 5}; // LES SOMMETS 2 ET 3 SONT PARTAGES PAR LES DEUX PATCHS.
	// LE PATCH DEFINI PAR LES SOMMETS 0, 1, 2 ET 3 EST DANS LA MOITIE DU BAS DE LA FENETRE
	// LE PATCH DEFINI PAR LES SOMMETS 0, 1, 2 ET 3 EST DANS LA MOITIE DU HAUT DE LA FENETRE

	glGenBuffers( 1, &MesSommetsID); // BUFFER DES SOMMETS
	glBindBuffer(GL_ARRAY_BUFFER, MesSommetsID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(MesSommets), (void*)MesSommets, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glGenBuffers( 1, &MesIndicesID); // BUFFER DES INDICES
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, MesIndicesID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(MesIndices), (void*)MesIndices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

}

void EncapsuleLesDonneesDansUnVao()
{
	glGenVertexArrays( 1, &MonVAO); // ON GENERE UN VAO
	glBindVertexArray(MonVAO);

	glEnableVertexAttribArray(0); // ON ACTIVE LE TABLEAU D'ATTRIBUT D'INDICE 0

        glBindBuffer(GL_ARRAY_BUFFER, MesSommetsID); // ON ACTIVE LE VBO DES SOMMETS
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, MesIndicesID); // ON ACTIVE LE VBO DES INDICES

       // ON INDIQUE COMMENT LES DONNEES SONT TRANSMISES AU PROGRAMME
	glVertexAttribPointer(
            0,                  // indice de l'attribut à configurer, correspond à la ligne "layout(location=0) in vec3 MonSommetEnEntree;" dans le vertex shader
            3,                  // nombre de composante par position. ici 3 composantes: (x, y, z).
            GL_FLOAT,           // type des composantes
            GL_FALSE,           // normalisation des composantes
            0,                  // decalage des composantes
            (void*)0            // offset des composantes
        );
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindVertexArray(0);
}



int main( int argc, char *argv[])
{
	GLFWwindow *MaFenetre = InitialiseUnContexteEtUneFenetreOpenGL();

	InitialisationDuProgrammeGPUEtDesLiensVersCeProgramme();
	glUseProgram(MonProgrammeGLSL);

	InitialisationDesDonnees();
	EncapsuleLesDonneesDansUnVao();


	std::cout << "MonProgrammeGLSL = " << MonProgrammeGLSL << std::endl;

	std::cout<<"MonVAO = "<<MonVAO<< std::endl;

	glClearColor( 0.0, 0.0, 0.0, 1.0f);

	glClearDepth(1.0f);

	glEnable(GL_DEPTH_TEST);

// INDIQUE LE NOMBRE DE POINTS A TRAITER PAR PATCH
// ICI ON FAIT UNE INTERPOLATION BILINEAIRE DES 4 SOMMETS
	glPatchParameteri(GL_PATCH_VERTICES, 4);

// LES LIGNES SUIVANTES PEUVENT PERMETTRE DE SE PASSER DE TESSELATION CONTROL SHADER DANS LES CAS SIMPLES
	GLfloat DefaultOuterLevel[]={4.,4.,4.,4.};
	GLfloat DefaultInnerLevel[]={4.,4.};
	glPatchParameterfv(GL_PATCH_DEFAULT_OUTER_LEVEL, DefaultOuterLevel);
	glPatchParameterfv(GL_PATCH_DEFAULT_INNER_LEVEL, DefaultInnerLevel);

	while (!glfwWindowShouldClose( MaFenetre)) {

		TraiteLesEvenementsClavier( MaFenetre);

		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		glUniform1f( FinesseDecoupageID, (GLfloat) (6.0f)); // FINESSE DE LA TESSELLATION

		glBindVertexArray( MonVAO);
		glDrawElements( GL_PATCHES, 8, GL_UNSIGNED_INT, 0); // DESSIN DE DEUX PATCHS DE 4 SOMMETS CHACUN
		glBindVertexArray( 0);

		glfwSwapBuffers(MaFenetre);

		glfwPollEvents();
	}

	glDeleteBuffers(1, &MesIndicesID);
	glDeleteBuffers(1, &MesSommetsID);
	glDeleteVertexArrays(1, &MonVAO);

	glfwTerminate();
}


GLFWwindow* InitialiseUnContexteEtUneFenetreOpenGL()
{
	GLFWwindow* MaFenetre;

	// INITIALISATION DE GLFW
	if( !glfwInit() )
	{
		std::cout << "Echec de l'initialisation de GLFW" << std::endl;
		exit(EXIT_FAILURE);
	}

//	glfwWindowHint(GLFW_SAMPLES, 4); // Anti Aliasing
	glfwWindowHint(GLFW_SAMPLES, 0); // pas d'Anti Aliasing
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4); // OpenGL 4.1
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 1);
	glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);


	// Ouverture d'une fenêtre en 1024x768
	// et creation d'un contexte OpenGL
	if( !(MaFenetre = glfwCreateWindow(TAILLE_HORIZONTALE, TAILLE_VERTICALE, "Polytech_RICM4", NULL,NULL)) )
	{
		std::cout << "Echec de l'ouverture de fenetre OpenGL" << std::endl;
		glfwTerminate();
		exit(EXIT_FAILURE);
	}
	glfwMakeContextCurrent(MaFenetre);

	// Autorise GLFW a recevoir les appuis de touche
	glfwSetInputMode( MaFenetre, GLFW_STICKY_KEYS, GL_TRUE);

	// Initialisation de  GLEW
	glewExperimental = GL_TRUE; 
	if (glewInit() != GLEW_OK) {
		std::cout << "Echec de l'initialisation de GLEW" << std::endl;
		exit(EXIT_FAILURE);
	}

	// Affichage des donnees du contexte OpenGL
	const GLubyte* renderer = glGetString (GL_RENDERER);
	std::cout << "TYPE DE CARTE GRAPHIQUE : " << renderer << std::endl;

	const GLubyte* version = glGetString (GL_VERSION);
	std::cout << "VERSION DE PILOTE OpenGL : " << version << std::endl;

	const GLubyte* shading_version = glGetString (GL_SHADING_LANGUAGE_VERSION);
	std::cout << "VERSION GLSL: " << shading_version << std::endl;

	return MaFenetre;
}

void TraiteLesEvenementsClavier(GLFWwindow *MaFenetre)
{
	if (glfwGetKey( MaFenetre, GLFW_KEY_A ) == GLFW_PRESS) {
		std::cout << "Touche A\n";
	}
}


